@@ -31,12 +31,17 @@ def epd_convert(epd_dict):
 def read_project(datadir, load_sample_limit=20):
     '''
     Get all project data.
-    Return
-    ======
-    project_desc: dict of project meta data
-    cmap: composite map, ['_number_of_samples_', 'rt_length', 'dict_scan_rtime', 'list_mass_tracks', 'MassGrid']
-    epd: dict of empirical compounds
-    Ftable: pandas dataframe of feature table. Truncated if samples more than load_sample_limit.
+
+    Returns
+    -------
+    project_desc : 
+        dict of project meta data
+    cmap : 
+        composite map, ['_number_of_samples_', 'rt_length', 'dict_scan_rtime', 'list_mass_tracks', 'MassGrid']
+    epd : 
+        dict of empirical compounds
+    Ftable : 
+        pandas dataframe of feature table. Truncated if samples more than load_sample_limit.
     '''
     project_desc = json.load(open(os.path.join(datadir, 'project.json')))
     cmap = pickle.load( open(os.path.join(datadir, 'export', 'cmap.pickle'), 'rb') )
@@ -54,7 +59,7 @@ def read_project(datadir, load_sample_limit=20):
 
 def plot_xic(xics, mz_dict, track_id):
     '''
-    track_id is a str
+    Generates scatter plot for a mass track as dataframe. Test function.
     '''
     title = "Mass track viewer - m/z %4.4f" %(mz_dict[track_id])
     plot = xics.hvplot.scatter(x='rt', y=track_id, height=400, title=title,)
@@ -64,6 +69,27 @@ def plot_xic(xics, mz_dict, track_id):
 def cmapplot_mass_tracks(cmap, rt_list, color, track_id_number):
     '''
     return a hv plot of mass track, by track_id_number
+    
+    Parameters
+    ----------
+    cmap : 
+        CMAP as from ext_Experiment {
+            '_number_of_samples_': self.CMAP._number_of_samples_,
+            'rt_length': self.CMAP.rt_length,
+            'dict_scan_rtime': self.CMAP.dict_scan_rtime,
+            'list_mass_tracks': self.CMAP.composite_mass_tracks,
+            'MassGrid': dict(self.CMAP.MassGrid),
+        }
+    rt_list : 
+        list of retention time
+    color : 
+        color for scatter
+    track_id_number : 
+        index for a mass track in cmap['list_mass_tracks']
+
+    Returns
+    -------
+    A holoviz plot object.
     '''
     track_id_number = int(track_id_number)
     mz = cmap['list_mass_tracks'][track_id_number]['mz']
@@ -78,8 +104,9 @@ def cmapplot_mass_tracks(cmap, rt_list, color, track_id_number):
 
 def convert_dict_html(d, title=''):
     '''
-    Convert peak dictionary into readable HTML.
+    Convert a peak dictionary into readable HTML.
     May need to improve error handling since KeyError is potential problem.
+    Returns HTML as a string.
     '''
     s = title
     info = [ ('id_number: ', d['id_number'], ' - ', 'parent_masstrack_id: ', d['parent_masstrack_id'], ' - ', 'parent_epd_id: ', d.get('parent_epd_id', '')),
@@ -93,6 +120,9 @@ def convert_dict_html(d, title=''):
 
 
 def convert_dict_markdown(d, title=''):
+    '''
+    Convert a peak dictionary into Markdown string.
+    '''
     s = title
     for k,v in d.items():
         if k not in ['apex', 'left_base', 'right_base']:
@@ -100,6 +130,15 @@ def convert_dict_markdown(d, title=''):
     return s + '\n'
 
 def track_to_peaks(peakDict):
+    '''
+    Parameters
+    -----
+    peakDict, peak dictionary using 'id_number' as keys.
+    
+    Returns
+    -------
+    a dictionary using parent_masstrack_id as keys and ['id_number'] as values.
+    '''
     t = {}
     for P in peakDict.values():
         mid = str(P["parent_masstrack_id"])
@@ -121,15 +160,54 @@ def find_track_by_mz(cmap, rt_list, mz):
         return sorted(L)[0][1]
 
 def find_a_good_peak(peakDict):
-    '''find a good example feature/peak'''
+    '''
+    To find a good example feature/peak, which can be used as the feature on landing page.
+    '''
     good = [P for P in peakDict.values() if P['goodness_fitting'] > 0.9 and P['cSelectivity'] > 0.9]
     return good[0]
 
+def prepare_rt_alignment(cmap):
+    """this method prepares necessary dataframe for retention time alignment.
+
+    Parameters
+    ----------
+    cmap : dict
+        pickle information read include all we need for retention time alignment
+
+    Returns
+    -------
+    dataframe
+        pandas dataframe ready for holoview drawing
+     
+    """
+    sample_reference_list = [rt_record['reverse_rt_cal_dict'] for rt_record in cmap['rt_records']]
+    rt_length = cmap['rt_length']
+
+    for i, d in enumerate(sample_reference_list):
+        # imputate the equal key-value pair
+        missing_keys = {k: k for k in range(rt_length) if k not in d}
+        d.update(missing_keys)
+        
+        # sort the dicts by key & map scan number to rtime & calculate the deviation
+        deviation_dict = {cmap['dict_scan_rtime'][k]: cmap['dict_scan_rtime'][d[k]]-cmap['dict_scan_rtime'][k] for k in sorted(d)}
+
+        sample_reference_list[i] = deviation_dict
+
+    sample_name_list = [record['name'] for record in cmap['rt_records']]
+    deviation_df = pd.DataFrame(sample_reference_list).T
+    deviation_df.reset_index().rename(columns={'index':'rtime'})
+    # rename columns
+    deviation_df.columns = sample_name_list
+
+    return deviation_df
 
 #
 # Summary panel
 #
-def get_summary_panel(project_desc, peakDict, epdDict, Ftable):
+def get_summary_panel(project_desc, peakDict, epdDict, Ftable, cmap):
+    '''
+    Get a summary panel, returns a panel.Column of multiple tabs for summary metrics.
+    '''
     desc0 = "Project retrieved from %s, %d features and %d empirical compounds." %(project_desc['outdir'],
                     len(peakDict), len(epdDict)
                     )
@@ -137,38 +215,48 @@ def get_summary_panel(project_desc, peakDict, epdDict, Ftable):
             "</ul><ul>This Dashboard shows all features of varying measurement quality. We recommend to use the preferred feature table for data analysis.</ul>"
     
     description = pn.pane.HTML(html)
-    height = 240
+    height = 330
+    width = 1000
     num_bins = int(np.sqrt(Ftable.shape[0]))
     # Feature m/z distribution histogram
     mz_distribution = Ftable['mz'].hvplot.hist(bins=num_bins, tools=[], 
-                        height=height, alpha=0.7, color='gold', hover=False,
+                        width = width, height=height, alpha=0.7, color='gold', hover=False,
                         ylabel="Count", title="Feature distribution by m/z"
                         ).opts(toolbar=None)
     # eature distribution by RT scatterplot
     Ftable['peak_area_sqrt'] = np.sqrt(Ftable['peak_area'])
-    rt_distribution = Ftable.hvplot.scatter(x='rtime', y='peak_area_sqrt', tools=[], height=height, size=2, hover=False,
+    rt_distribution = Ftable.hvplot.scatter(x='rtime', y='peak_area_sqrt', tools=[], width=width, height=height, size=2, hover=False,
                         title="Feature distribution by RT").opts(toolbar=None)
     # quality metrics
     Ftable['log10(SNR)'] = np.log10(Ftable['snr']+1)
     SNR = Ftable['log10(SNR)'].hvplot.hist(bins=num_bins, tools=[], 
-                        height=height, alpha=0.7, color='green', hover=False, 
+                        width=width,height=height, alpha=0.7, color='green', hover=False, 
                         ylabel="Count", title="Signal to noise ratio"
                         ).opts(toolbar=None)
     PeakShape = Ftable['goodness_fitting'].hvplot.hist(bins=num_bins, tools=[], 
-                        height=height, alpha=0.7, color='grey', hover=False, 
+                        width=width,height=height, alpha=0.7, color='grey', hover=False, 
                         ylabel="Count", title="Peak shape"
                         ).opts(toolbar=None)
     cSel = Ftable['cSelectivity'].hvplot.hist(bins=num_bins, tools=[], 
-                        height=height, alpha=0.7, color='red', hover=False, 
+                        width=width,height=height, alpha=0.7, color='red', hover=False, 
                         ylabel="Count", title="Chromatographic Selectivity"
                         ).opts(toolbar=None)
+    
+    # hard cap for legend is 20
+    if_legend = cmap['_number_of_samples_'] <= 20
+    RTAlign = prepare_rt_alignment(cmap).hvplot.line(title="Retention Time Deviation vs Retention Time",xlabel="Retention Time (sec)", 
+                ylabel="Retention Time Deviation (sec)", width=width, height=height,hover=False).opts(
+                show_legend=if_legend, legend_opts=dict(
+                title='',label_text_font_size='8pt',spacing=-5, location=(5, (height/2-cmap['_number_of_samples_']*11)), padding=3), toolbar=None)
+    
 
     feature_distribution = pn.Tabs(
-        ("Feature m/z distribution histogram", mz_distribution),
-        ("Feature distribution by RT scatterplot", rt_distribution),
-        ("Signal to noise ratio", SNR),
-        ("Peak shape", PeakShape),
-        ("Chromatographic Selectivity", cSel),
+        ("Feature m/z distribution histogram", pn.Column(mz_distribution)),
+        ("Feature distribution by RT scatterplot", pn.Column(rt_distribution)),
+        ("Signal to noise ratio", pn.Column(SNR)),
+        ("Peak shape", pn.Column(PeakShape)),
+        ("Chromatographic Selectivity", pn.Column(cSel)),
+        ("Retention Time Alignment", pn.Column(RTAlign)),
         )
     return pn.Column(
                     description,
@@ -188,7 +276,7 @@ def dashboard(project_desc, cmap, epd, Ftable):
     a_good_peak = find_a_good_peak(peakDict)
     track2peaks = track_to_peaks(peakDict)
     rt_list = [cmap['dict_scan_rtime'][ii] for ii in range(cmap['rt_length'])]
-    summary = get_summary_panel(project_desc, peakDict, epdDict, Ftable)
+    summary = get_summary_panel(project_desc, peakDict, epdDict, Ftable, cmap)
 
     #
     # Feature browser
