@@ -7,14 +7,19 @@ from mass2chem.search import *
 # jms-metabolite-services
 from jms.dbStructures import knownCompoundDatabase, ExperimentalEcpdDatabase
 
-# from .samples import SimpleSample
-from .mass_functions import complete_mass_paired_mapping
+from .default_parameters import adduct_search_patterns, \
+                                adduct_search_patterns_neg, \
+                                    isotope_search_patterns, \
+                                        extended_adducts
+
+from .mass_functions import all_mass_paired_mapping
 from .constructors import CompositeMap
 from .json_encoder import NpEncoder
 
 try:
     import importlib.resources as pkg_resources
 except ImportError:
+    # Shuzhao - what is the purpose of this? I'm assuming this is vestigial and can be removed?
     import importlib_resources as pkg_resources
 
 from . import db
@@ -27,16 +32,31 @@ if not sys.warnoptions:
 class ext_Experiment:
     '''
     Similar to metDataModel.core.Experiment with preprocessing methods.
+
     This encapsulates a set of LC-MS files using the same experimental method 
     (chromatography and ionization) to be processed together.
+
     E.g., data from postive ESI and negative ESI should not in the same ext_Experiment instance.
+
+    This class has annotation and export functions.
+    
+    Default asari work flow is in `ext_Experiment.process_all`.
     '''
     def __init__(self, sample_registry, parameters):
         '''
         This is the overall container for all data in an experiment/project.
-        sample_registry: dictionary of sample and selected data after mass track extraction.
+
+        Parameters
+        ----------
+        sample_registry : dict
+            dictionary of sample and selected data after mass track extraction.
             The bulk mass track data are not kept in memory unless specified so.
-        parameters: processing parameters passed from main.py.
+        parameters : dict
+            processing parameters passed from main.py.
+
+        Updates
+        -------
+        Major class attributes including self.number_of_samples, number_scans, reference_sample_id.
         '''
         self.sample_registry = sample_registry
         self.valid_sample_ids = self.get_valid_sample_ids()
@@ -55,7 +75,8 @@ class ext_Experiment:
     def get_reference_sample_id(self):
         '''
         get_reference_sample_id either by user specification, or
-        as sample of most number_anchor_mz_pairs, limited to first 100 samples to search.
+        using the sample of most number_anchor_mz_pairs.
+        This assumes the sample of most good m/z values has a good coverage of features.
         '''
         if self.parameters['reference']:
             # match file name; k is sm['sample_id']
@@ -75,9 +96,22 @@ class ext_Experiment:
             return None
         
     def get_valid_sample_ids(self):
+        '''
+        Get valid sample ids, as some samples may not be extracted successfully.
+        '''
         return [k for k,v in self.sample_registry.items() if v['status:eic'] == 'passed']
 
     def get_max_scan_number(self, sample_registry):
+        '''
+        Return max scan number among samples, or None if no valid sample.
+
+        Parameters
+        ----------
+        sample_registry: dict
+            a dict that maps sample IDs to sample data
+        '''
+
+        # todo - why does this function need to take sample_registry as an external argument vs. self.sample_registry?
         if sample_registry:
             return max([sample_registry[k]['max_scan_number'] for k in self.valid_sample_ids]) + 1
         else:
@@ -85,15 +119,20 @@ class ext_Experiment:
 
     def process_all(self):
         '''
-        This is default asari workflow.
+        This is the default asari workflow.
+        
         1. Build MassGrid, using either pairwise (small study) or clustering method. 
-            Choose one reference from all samples for the largest number of landmark m/z tracks.
+           Choose one reference from all samples for the largest number of landmark m/z tracks.
         2. RT alignment via a LOWESS function, using selective landmark peaks.
         3. Build composite elution profile (composite_mass_tracks),
-            by cumulative sum of mass tracks from all samples after RT correction.
+           by cumulative sum of mass tracks from all samples after RT correction.
         4. Global peak detection is performed on each composite massTrack.
         5. Mapping global peaks (i.e. features) back to all samples and extract sample specific peak areas.
-            This completes the FeatureTable.
+           This completes the FeatureTable.
+
+        Updates
+        -------
+        self.CMAP as instance of CompositeMap, and MassGrid, composite map and features within.
         '''
         self.CMAP = CompositeMap(self)
         self.CMAP.construct_mass_grid()
@@ -106,29 +145,67 @@ class ext_Experiment:
         '''
         Export all files.
         Annotation of features to empirical compounds is done here.
+
+        Parameters
+        ----------
+        anno: bool, optional, default: True
+            if true, generate annotation files, export CMAP pickle and do QC plot;
+            else skip annotating.
         '''
         self.CMAP.MassGrid.to_csv(
             os.path.join(self.parameters['outdir'], 'export', self.parameters['mass_grid_mapping']) )
         if anno:
+            for peak in self.CMAP.FeatureList:
+                peak['id'] = str(peak['id_number'])
             self.export_CMAP_pickle()
             self.annotate()
+            self.generate_qc_plot_pdf()
         self.export_feature_tables()
         self.export_log()
 
     def annotate(self):
         '''
+        Annotate features via JMS (jms.dbStructures) and khipu.
+        The pre-annotation step is khipu based emprical compound construction, followed by 
+        three steps of annotation:
+
+        1) Search known compound database, via neutral mass inferred by khipu
+
+        2) Search singletons for a formula match
+
+        3) Encapsulate remaining features in empCpd format, so that all are exported consistently.
+
         Export Feature_annotation as tsv, Annotated_empricalCompounds in both JSON and pickle.
-        Reference databases can be pre-loaded.
-        With db_mass_calibrate to theoretical values, 
+        Reference databases can be pre-loaded. 
+        Measured m/z values are calibrated to database based values (db_mass_calibrate).
+
+        Note
+        ----
+        This produces default annotation with asari, 
+        but one can redo annotation on the features afterwards, using a method of choice.
+        With JMS/khipu, one can also pass custom adduct/isotopes to EED.adduct_patterns etc. 
+        See ExperimentalEcpdDatabase.get_isotope_adduct_patterns().
         '''
         self.load_annotation_db()
         self.db_mass_calibrate()
 
-        EED = ExperimentalEcpdDatabase(mode=self.mode, mz_tolerance_ppm=self.mz_tolerance_ppm, rt_tolerance=2)
+        # asari uses seconds for rt
+        EED = ExperimentalEcpdDatabase(mode=self.mode, 
+                                       mz_tolerance_ppm=self.mz_tolerance_ppm, rt_tolerance=2)
+        # passing patterns from .default_parameters
+        if self.mode == 'pos':
+            EED.adduct_patterns = adduct_search_patterns
+        else:
+            EED.adduct_patterns = adduct_search_patterns_neg
+        EED.isotope_search_patterns = isotope_search_patterns
+        EED.extended_adducts = extended_adducts
+
         EED.build_from_list_peaks(self.CMAP.FeatureList)
+        # It takes three steps to take care of all features. First khipu organized empCpds
         EED.extend_empCpd_annotation(self.KCD)
+        # Second, singletons that get a formula match in KCD
         EED.annotate_singletons(self.KCD)       
-        # EED.dict_empCpds misses some features 
+        # Third, the remaining features unmatched to anything (orphans). Exported for potential downstream work.
         EED.dict_empCpds = self.append_orphans_to_epmCpds(EED.dict_empCpds)
 
         self.export_peak_annotation(EED.dict_empCpds, self.KCD, 'Feature_annotation')
@@ -146,13 +223,58 @@ class ext_Experiment:
         with open(outfile, 'wb') as f:
             pickle.dump(EED.dict_empCpds, f, pickle.HIGHEST_PROTOCOL)
 
+    def generate_qc_plot_pdf(self, outfile="qc_plot.pdf"):
+        '''
+        Generates a PDF figure of a combined plot feature quality metrics.
+        Used only when --anno True (default).
+        Skip if matplotlib is missing.
+        '''
+        try:
+            from .qc import asari_qc_plot
+            # using prefilter CMAP.FeatureTable
+            asari_qc_plot(self.CMAP.FeatureTable,
+                          outfile=os.path.join(self.parameters['outdir'], 'export', outfile))
+
+        except ImportError:
+            print("[QC plot] cannot import matplotlib, skipping.")
+
     def export_CMAP_pickle(self):
         '''
         Export main CMAP data and MassGrid to pickle, which can be used for visual data exploration.
+
+        Included in exported pickle:{
+            '_number_of_samples_': self.CMAP._number_of_samples_,
+            'rt_length': self.CMAP.rt_length,
+            'rt_reference_landmarks': [p['apex'] 
+                                       for p in self.CMAP.good_reference_landmark_peaks],
+            'rt_records': [sample.get_rt_calibration_records()
+                                        for sample in self.all_samples
+                                        ],
+            'dict_scan_rtime': self.CMAP.dict_scan_rtime,
+            'list_mass_tracks': self.CMAP.composite_mass_tracks,
+            'MassGrid': dict(self.CMAP.MassGrid),}
+
+        rt_records includes for each sample: {
+            'sample_id': self.sample_id,
+            'name': self.name,
+            'rt_landmarks': self.rt_landmarks,
+            'reverse_rt_cal_dict': self.reverse_rt_cal_dict,
+        }
+
+        Note
+        ----
+            RT calibration is exported to include sample.reverse_rt_cal_dict, 
+            i.e. {key=reference scan number, value=sample specific scan number}.
+            May add data throttle in the future. The file cmap.pickle can get big.
         '''
         _export = {
             '_number_of_samples_': self.CMAP._number_of_samples_,
             'rt_length': self.CMAP.rt_length,
+            'rt_reference_landmarks': [p['apex'] 
+                                       for p in self.CMAP.good_reference_landmark_peaks],
+            'rt_records': [sample.get_rt_calibration_records()
+                                        for sample in self.all_samples
+                                        ],
             'dict_scan_rtime': self.CMAP.dict_scan_rtime,
             'list_mass_tracks': self.CMAP.composite_mass_tracks,
             'MassGrid': dict(self.CMAP.MassGrid),
@@ -163,36 +285,67 @@ class ext_Experiment:
 
     def load_annotation_db(self, src='hmdb4'):
         '''
-        Database of known compound using JMS. Will add more options later.
+        Load database of known compound using jms.dbStructures.knownCompoundDatabase.
+        The compound tree is precomputed indexing.
+        The `src` parameter is not used now, but placeholder to add more options later.
+
+        Parameters
+        ----------
+        src: str, optional, default: hmdb4
+            not used but can, in the future, dictate which database is used to generate annotations
         '''
         self.KCD = knownCompoundDatabase()
-        self.KCD.mass_indexed_compounds = pickle.load( pkg_resources.open_binary(db, 'mass_indexed_compounds.pickle') )
-        self.KCD.emp_cpds_trees = pickle.load( pkg_resources.open_binary(db, 'emp_cpds_trees.pickle') )
+        self.KCD.mass_indexed_compounds = pickle.load( 
+            pkg_resources.open_binary(db, 'mass_indexed_compounds.pickle') )
+        self.KCD.emp_cpds_trees = pickle.load( 
+            pkg_resources.open_binary(db, 'emp_cpds_trees.pickle') )
 
     def db_mass_calibrate(self, required_calibrate_threshold=0.000002):
         '''
-        Use KCD.evaluate_mass_accuracy_ratio to check systematic mass shift.
+        Use KCD.evaluate_mass_accuracy_ratio to check systematic mass shift,
+        which is calculated as the average ppm difference between measured m/z and theoretical values.
         If greater than required_calibrate_threshold (default 2 ppm), 
         calibrate m/z values for the whole experiment by updating self.CMAP.FeatureList.
 
-        good_reference_landmark_peaks: [{'ref_id_num': 99, 'apex': 211, 'height': 999999}, ...]
-        ref_id_num -> index number of mass track in MassGrid.
+        Parameters
+        ---------
+        required_calibrate_treshold: float, optional, default: 0.000002
+            if the mass shift exceeds this value, mass correction will be applied. 
+
+        Note
+        ----
+        Data format in good_reference_landmark_peaks: 
+        [{'ref_id_num': 99, 'apex': 211, 'height': 999999}, ...],
+        where ref_id_num is index number of mass track in MassGrid.
         '''
-        mz_landmarks = [self.CMAP.MassGrid['mz'][p['ref_id_num']] for p in self.CMAP.good_reference_landmark_peaks]
-        mass_accuracy_ratio = self.KCD.evaluate_mass_accuracy_ratio(mz_landmarks, mode=self.mode, mz_tolerance_ppm=10)
+        mz_landmarks = [self.CMAP.MassGrid['mz'][p['ref_id_num']] 
+                        for p in self.CMAP.good_reference_landmark_peaks]
+        mass_accuracy_ratio = self.KCD.evaluate_mass_accuracy_ratio(
+            mz_landmarks, mode=self.mode, mz_tolerance_ppm=10)
         if mass_accuracy_ratio:
             if abs(mass_accuracy_ratio) > required_calibrate_threshold:
-                print("Mass shift is greater than %2.1f ppm. Correction applied." %(required_calibrate_threshold*1000000))
+                print("Mass shift is greater than %2.1f ppm. Correction applied." 
+                      %(required_calibrate_threshold*1000000))
                 _correction = mass_accuracy_ratio + 1
                 for F in self.CMAP.FeatureList:
                     F['mz'] = F['mz'] / _correction
                     F['mz_corrected_by_division'] = _correction
         else:
-            print("Mass accuracy check is skipped, too few mz_landmarks (%d) matched." %len(mz_landmarks))
+            print("Mass accuracy check is skipped, too few mz_landmarks (%d) matched." 
+                  %len(mz_landmarks))
 
     def append_orphans_to_epmCpds(self, dict_empCpds):
         '''
-        # EED.dict_empCpds does not include features without formula match, and we add them there.
+        This is the third step of feature annotation in self.annotate,
+        to encapsulate features without annotation in empCpd format.
+        Input via dict_empCpds and returns updated dict_empCpds.
+        See also: annotate
+
+        Parameters
+        ----------
+        dict_empCpds: dict
+            a dictionary of empirical compounds in empCpd format
+
         '''
         all_feature_ids = []
         for _, V in dict_empCpds.items():
@@ -211,12 +364,20 @@ class ext_Experiment:
 
     def export_peak_annotation(self, dict_empCpds, KCD, export_file_name_prefix):
         '''
-        Export feature annotation.
-        interim_id is empCpd id. dict_empCpds as seen in JMS.
+        Export feature annotation to tab delimited tsv file, where interim_id is empCpd id.
         
+        Parameters
+        ----------
+        dict_empCpds : dict
+            dictionary of empirical compounds, using interim_id as key, as seen in JMS.
+        KCD : KnownCompoundDatabase instance
+            the known compound database that was used in annotating the empirical compounds.
+        export_file_name_prefix : str
+            to used in output file name.
         '''
-        s = "[peak]id_number\tmz\trtime\tapex(scan number)\t[EmpCpd]interim_id\t[EmpCpd]ion_relation\tneutral_formula\tneutral_formula_mass\
-        \tname_1st_guess\tmatched_DB_shorts\tmatched_DB_records\n"
+        s = "[peak]id_number\tmz\trtime\tapex(scan number)\t[EmpCpd]interim_id\
+            \t[EmpCpd]ion_relation\tneutral_formula\tneutral_formula_mass\
+            \tname_1st_guess\tmatched_DB_shorts\tmatched_DB_records\n"
         
         for _, V in dict_empCpds.items():
             name_1st_guess, matched_DB_shorts, matched_DB_records = '', '', ''
@@ -229,7 +390,8 @@ class ext_Experiment:
 
             for peak in V['MS1_pseudo_Spectra']:
                 s += '\t'.join([str(x) for x in [
-                    peak['id_number'], peak['mz'], peak['rtime'], peak['apex'], V['interim_id'], peak.get('ion_relation', ''),
+                    peak['id_number'], peak['mz'], peak['rtime'], peak['apex'], 
+                    V['interim_id'], peak.get('ion_relation', ''),
                     V['neutral_formula'], V['neutral_formula_mass'],
                     name_1st_guess, matched_DB_shorts, matched_DB_records]]) + "\n"
 
@@ -242,6 +404,13 @@ class ext_Experiment:
     def select_unique_compound_features(self, dict_empCpds):
         '''
         Get unique feature by highest composite peak area per empirical compound. 
+        One may consider alternatives to select the peak representing an empirical compound,
+        e.g. by SNR or M+H, M-H ions. This is can be done separately on the exported files.
+
+        Parameters
+        ----------
+        dict_empCpds : dict
+            dictionary of empirical compounds, using interim_id as key, as seen in JMS.
         '''
         self.selected_unique_features = {}
         for interim_id, V in dict_empCpds.items():
@@ -252,7 +421,8 @@ class ext_Experiment:
                 )
             else:
                 try:
-                    all_peaks = sorted([(peak['peak_area'], peak['goodness_fitting'], peak) for peak in V['MS1_pseudo_Spectra']],
+                    all_peaks = sorted([(peak['peak_area'], peak['goodness_fitting'], peak) 
+                                        for peak in V['MS1_pseudo_Spectra']],
                                         reverse=True)
                     best_peak = all_peaks[0][2]
                 except TypeError:
@@ -262,15 +432,31 @@ class ext_Experiment:
                     interim_id, V['neutral_formula'], best_peak.get('ion_relation', '')
                 )
 
-    def export_feature_tables(self, _snr=10, _peak_shape=0.7, _cSelectivity=0.7):
+    def export_feature_tables(self, _snr=2, _peak_shape=0.7, _cSelectivity=0.7):
         '''
-        To export features tables:
-        1) preferred table under `outdir`, after quality filtering by SNR, peak shape and chromatographic selectivity.
-        2) full table under `outdir/export/`
-        3) unique compound table under `outdir/export/`
-        4) dependent on `target` extract option, a targeted_extraction table under `outdir`.
-        Filtering parameters (_snr, _peak_shape, _cSelectivity) only apply to preferred table and unique_compound_table.
-        Full table is filtered by initial peak detection parameters of lower snr and gaussian_shape.
+        To export multiple features tables. 
+        Filtering parameters (_snr, _peak_shape, _cSelectivity) only 
+        apply to preferred table and unique_compound_table.
+        Full table is filtered by initial peak detection parameters, 
+        which contains lower values of snr and gaussian_shape.
+
+        Parameters
+        ----------
+        _snr: float, optional, default: 2
+            signal to noise ratio, peaks must have SNR above this value to be a preferred feature
+        _peak_shape: float, optional, default: 0.7
+            the goodness fitting for a peak must be above this value to be a preferred feature
+        _cSelectivity: float, optional, default: 0.7
+            the cSelectivity of a peak must be above this value to be a preferred feature
+
+        Outputs
+        -------
+        Multiple features tables under output directory:
+        1. preferred table under `outdir`, after quality filtering 
+           by SNR, peak shape and chromatographic selectivity.
+        2. full table under `outdir/export/`
+        3. unique compound table under `outdir/export/`
+        4. dependent on `target` extract option, a targeted_extraction table under `outdir`. 
         '''
         good_samples = [sample.name for sample in self.all_samples] 
         filtered_FeatureTable = self.CMAP.FeatureTable[good_samples]                       
@@ -295,7 +481,7 @@ class ext_Experiment:
 
         # extract targeted m/z features
         if 'target' in self.parameters and self.parameters['target']:  
-            matched_list, _, target_unmapped = complete_mass_paired_mapping(
+            matched_list, _, target_unmapped = all_mass_paired_mapping(
                 filtered_FeatureTable['mz'].to_list(), self.parameters['target'], self.parameters['mz_tolerance_ppm']
             )
             print("\nIn targeted extraction, %d target mz values are not found in this dataset: " %len(target_unmapped))
