@@ -3,7 +3,7 @@ from yaml import load, Loader
 
 from asari import __version__
 from .workflow import *
-from .defaul_parameters import PARAMETERS
+from .default_parameters import PARAMETERS
 
 booleandict = {'T': True, 'F': False, 1: True, 0: False, 
                    'True': True, 'False': False, 'TRUE': True, 'FALSE': False, 'true': True, 'false': False,
@@ -15,15 +15,23 @@ def main(parameters=PARAMETERS):
     '''
     asari, Trackable and scalable Python program for high-resolution LC-MS metabolomics data preprocessing.
 
-        analyze: analyze a single mzML file to print summary of statistics and recommended parameters.
-        process: LC-MS data preprocessing
-        xic: construct mass trakcs (chromatogram) from mzML files
-        extract: targeted extraction of given m/z list
-        annotate: annotate a list of features
-        join: merge multiple processed projects (possibly split a large dataset)
-        viz: start interactive data visualization and exploration.
+        * analyze: analyze a single mzML file to print summary of statistics and recommended parameters.
+        * process: LC-MS data preprocessing
+        * xic: construct mass trakcs (chromatogram) from mzML files
+        * extract: targeted extraction of given m/z list
+        * annotate: annotate a list of features
+        * join: merge multiple processed projects (possibly split a large dataset)
+        * viz: start interactive data visualization and exploration.
+
+    Parameters
+    ----------
+    parameters : dict
+        This dictionary contains a number of key value pairs that determine the behavior of various aspects
+        of the asari processing. The parameters can be seen in default_parameters.py. Command line arguments
+        will override any defaults and any values provided in the parameters.json file. 
     '''
     def __run_process__(parameters, args):
+        
         # main process function
         list_input_files = read_project_dir(args.input)
         if not list_input_files:
@@ -69,12 +77,14 @@ def main(parameters=PARAMETERS):
 
     parser.add_argument('--autoheight', default=False,
             help='automatic determining min peak height')
-    parser.add_argument('--reverse_detection', default=False,
-            help='the reverse peak detection algorithm; not used now')
+    parser.add_argument('--peak_area', default='sum',
+            help='peak area culculation, sum, auc or gauss for area under the curve')
     parser.add_argument('--pickle', default=False, 
-            help='keep all intermediate pickle files')
+            help='keep all intermediate pickle files, ondisk mode only.')
     parser.add_argument('--anno', default=True, 
             help='perform default annotation after processing data')
+    parser.add_argument('--debug', default=False, 
+            help='Toggle on debug mode')
 
     args = parser.parse_args()
 
@@ -85,6 +95,7 @@ def main(parameters=PARAMETERS):
         )
     parameters['multicores'] = min(mp.cpu_count(), parameters['multicores'])
     parameters['input'] = args.input
+    parameters['debug'] = booleandict[args.debug]
     
     if args.mode:
         parameters['mode'] = args.mode
@@ -96,8 +107,8 @@ def main(parameters=PARAMETERS):
         parameters['project_name'] = args.project
     if args.output:
         parameters['outdir'] = args.output
-    if args.reverse_detection:
-        parameters['reverse_detection'] = booleandict[args.reverse_detection]
+    if args.peak_area:
+        parameters['peak_area'] = args.peak_area
     if args.pickle:
         parameters['pickle'] = booleandict[args.pickle]
     if args.anno:
@@ -128,7 +139,7 @@ def main(parameters=PARAMETERS):
     elif args.run == 'annotate':
         # Annotate a user supplied feature table
         from .annotate_user_table import annotate_user_featuretable
-        annotate_user_featuretable(args.input, parameters=parameters)
+        annotate_user_featuretable(args.input, parameters=parameters, rtime_tolerance=2)
 
     elif args.run == 'join':
         # input a list of directoreis, each a result of asari process
