@@ -20,9 +20,38 @@ from .mass_functions import *
 # -----------------------------------------------------------------------------
 def process_project(list_input_files, parameters):
     '''
-    list_input_files: mzML files.
-    parameters: dictionary of parameters.
-    Will adjust disk writing according to database_mode
+    This defines the main work flow in processing a list of LC-MS files, 
+    creates the output folder with a time stamp, and uses sample registry to coordinate parallel processing.
+    The whole project data are tracked in experiment.ext_Experiment class.
+
+    Parameters
+    ----------
+    list_input_files : list[str]
+        list of centroided mzML filepaths from LC-MS metabolomics. Usually found in a folder.
+    parameters : dict
+        parameter dictionary passed from main.py, 
+        which imports from default_parameters and updates the dict by user arguments.
+
+    Outputs
+    -------
+    A local folder with asari processing result, e.g::
+    
+        rsvstudy_asari_project_427105156
+        ├── Annotated_empricalCompounds.json
+        ├── Feature_annotation.tsv
+        ├── export
+        │   ├── _mass_grid_mapping.csv
+        │   ├── cmap.pickle
+        │   ├── full_Feature_table.tsv
+        │   └── unique_compound__Feature_table.tsv
+        ├── pickle
+        │   ├── Blank_20210803_003.pickle
+        │   ├── ...
+        ├── preferred_Feature_table.tsv
+        └── project.json
+        
+
+    The pickle folder is removed after the processing by default.
     '''
     sample_registry = register_samples(list_input_files)
     if parameters['database_mode'] == 'auto':
@@ -59,6 +88,19 @@ def process_project(list_input_files, parameters):
 def read_project_dir(directory, file_pattern='.mzML'):
     '''
     This reads centroided LC-MS files from directory.
+    Returns a list of files that match file_pattern.
+
+    Parameters
+    ----------
+    directory: str
+        path to a directory containing mzML files
+    file_pattern: str, optional, default: '.mzML'
+        files with this substring will be ingested
+
+    Return
+    ------
+    list of paths to files containing the file_pattern
+
     '''
     print("Working on ~~ %s ~~ \n\n" %directory)
     return [os.path.join(directory, f) for f in os.listdir(directory) if file_pattern in f]
@@ -66,6 +108,15 @@ def read_project_dir(directory, file_pattern='.mzML'):
 def register_samples(list_input_files):
     '''
     Establish sample_id here, return sample_registry as a dictionary.
+
+    Parameters
+    ---------
+    list_input_files: list[str]
+        list of input filepaths, each representing a sample
+
+    Return
+    ------
+    sample_registry, a dictionary of integer sample id's to filepaths
     '''
     sample_registry = {}
     ii = 0
@@ -75,6 +126,17 @@ def register_samples(list_input_files):
     return sample_registry
 
 def create_export_folders(parameters, time_stamp):
+    '''
+    Creates local directory for storing temporary files and output result.
+    A time stamp is added to directory name to avoid overwriting existing projects.
+
+    Parameters
+    ----------
+    paramaters: dict
+        passed from main.py to get outdir and project_name
+    time_stamp: str
+        a time_stamp string to prevent overwriting existing projects
+    '''
     parameters['outdir'] = '_'.join([parameters['outdir'], parameters['project_name'], time_stamp]) 
     parameters['tmp_pickle_dir'] = os.path.join(parameters['outdir'], 'pickle')
     os.mkdir(parameters['outdir'])
@@ -84,6 +146,11 @@ def create_export_folders(parameters, time_stamp):
 def remove_intermediate_pickles(parameters):
     '''
     Remove all temporary files under pickle/ to free up disk space. 
+
+    Parameters
+    ----------
+    paramaters: dict
+        passed from main.py to get tmp_pickle_dir
     '''
     print("Removing temporary pickle files...")
     for f in os.listdir(parameters['tmp_pickle_dir']):
@@ -100,8 +167,23 @@ def remove_intermediate_pickles(parameters):
 def make_iter_parameters(sample_registry, parameters, shared_dict):
     '''
     Generate iterables for multiprocess.starmap for getting sample mass tracks.
-    return:
-    [('sample_id', input_file, mode, mz_tolerance_ppm, min_intensity, min_timepoints, min_peak_height, output_file, shared_dict), ...]
+
+    Parameters
+    ----------
+    sample_registry : dict
+        dictionary like {'sample_id': ii, 'input_file': file}, 
+        generated by register_samples.
+    parameters : dict
+        parameter dictionary passed from main.py, 
+        which imports from default_parameters and updates the dict by user arguments.
+    shared_dict : dict
+        dictionary object used to pass data btw multiple processing.
+
+    Returns
+    -------
+    A list of iterative parameters, e.g.
+    [('sample_id', input_file, mode, mz_tolerance_ppm, min_intensity, min_timepoints, 
+    min_peak_height, output_file, shared_dict), ...]
     '''
     iters = []
     mz_tolerance_ppm = parameters['mz_tolerance_ppm']
@@ -109,7 +191,8 @@ def make_iter_parameters(sample_registry, parameters, shared_dict):
     min_timepoints = parameters['min_timepoints']
     min_peak_height = parameters['min_peak_height']
     for sample in sample_registry.values():
-        outfile = os.path.join(parameters['outdir'], 'pickle', os.path.basename(sample['input_file']).replace('.mzML', '')+'.pickle')
+        outfile = os.path.join(parameters['outdir'], 'pickle', 
+                               os.path.basename(sample['input_file']).replace('.mzML', '')+'.pickle')
         iters.append(
             (sample['sample_id'], sample['input_file'], parameters['mode'], parameters['database_mode'],
             mz_tolerance_ppm, min_intensity, min_timepoints, min_peak_height, outfile,
@@ -120,8 +203,25 @@ def make_iter_parameters(sample_registry, parameters, shared_dict):
 
 def batch_EIC_from_samples_(sample_registry, parameters):
     '''
-    multiprocessing of mass track extraction of samples, and return shared_dict.
-    More anchors mean better coverage of features, helpful to select reference sample.
+    Batch extraction of mass tracks from samples via multiprocessing.
+    
+    Parameters
+    ----------
+    sample_registry : dict
+        dictionary like {'sample_id': ii, 'input_file': file}, 
+        generated by register_samples.
+    parameters : dict
+        parameter dictionary passed from main.py, 
+        which imports from default_parameters and updates the dict by user arguments.
+
+    Returns
+    -------
+    shared_dict : dict
+        dictionary object used to pass data btw multiple processing.
+
+    See also
+    --------
+    single_sample_EICs_
     '''
     with mp.Manager() as manager:
         shared_dict = manager.dict()
@@ -137,15 +237,54 @@ def single_sample_EICs_(sample_id, infile, ion_mode, database_mode,
                     mz_tolerance_ppm, min_intensity, min_timepoints, min_peak_height, outfile, 
                     shared_dict):
     '''
-    Process infile. 
+    Extraction of mass tracks from a single sample. Used by multiprocessing in batch_EIC_from_samples_.
     `shared_dict` is used to pass back information, thus critical. Designed here as
-    sample_id: ('status:mzml_parsing', 'status:eic', outfile,
-                max_scan_number, list_scan_numbers, list_retention_time,
-                track_mzs,
-                number_anchor_mz_pairs, anchor_mz_pairs, 
-                dict({mass tracks}) )
+    sample_id: 
+        ('status:mzml_parsing', 'status:eic', outfile,
+        max_scan_number, list_scan_numbers, list_retention_time,
+        track_mzs,
+        number_anchor_mz_pairs, anchor_mz_pairs, 
+        dict({mass tracks}) )
+
     track_mzs or anchor_mz_pairs are used later for aligning m/z tracks.
+
     list of scans starts from 0. 
+    
+    `anchor_mz_pairs` are defined as m/z pairs that match to 13C/12C pattern.
+    More anchors mean better coverage of features, helpful to select reference sample.
+
+    Parameters
+    ----------
+    sample_id : int
+        sample id passed from sample_registry by make_iter_parameters.
+    infile : str
+        input mzML filepath, passed from sample_registry by make_iter_parameters.
+    ion_mode: str
+        from parameter dictionary, marks if acquisition was in positive or negative mode 
+    database_mode: str
+        from parameter dictionary, marks if intermediatesare kept on disk or in memory
+    mz_tolerance_ppm: float
+        from parameter dictionray, the assumed mz resolution of the instrument
+    min_intensity: float
+        peaks below this value are ignored
+    min_timepoints: int
+        then number of time points a peak must span to be considered a peak
+    min_peak_height: float
+        peaks below this height are ignored
+    outfile : str
+        where the output will be written
+        passed from parameter dictionary by make_iter_parameters.
+    shared_dict : dict
+        dictionary object used to pass data btw multiple processing.
+
+    Updates
+    -------
+    shared_dict : dict
+        dictionary object used to pass data btw multiple processing.
+
+    See also
+    --------
+    batch_EIC_from_samples_, make_iter_parameters
     '''
     new = {'sample_id': sample_id, 'input_file': infile, 'ion_mode': ion_mode,}
     list_mass_tracks = []
@@ -164,10 +303,10 @@ def single_sample_EICs_(sample_id, infile, ion_mode, database_mode,
             list_mass_tracks.append( {
                 'id_number': ii, 
                 'mz': track[0],
-                # 'rt_scan_numbers': track[1],              # format changed after v1.5
+                # 'rt_scan_numbers': track[1],       # format changed after v1.5
                 'intensity': track[1], 
                 } )
-            track_mzs.append( (track[0], ii) )                  # keep a reconrd in sample registry for fast MassGrid align
+            track_mzs.append( (track[0], ii) )       # keep a reconrd in sample registry for fast MassGrid align
             ii += 1
 
         new['list_mass_tracks'] = list_mass_tracks
@@ -208,9 +347,22 @@ def single_sample_EICs_(sample_id, infile, ion_mode, database_mode,
 
 def process_xics(list_input_files, parameters):
     '''
-    Get XICs (mass tracks) from a folder of centroid mzML files.
-    list_input_files: mzML files.
-    parameters: dictionary of most parameters.
+    Get XICs (aka EICs or mass tracks) from a folder of centroid mzML files 
+    and store in local pickle files.
+
+    Parameters
+    ----------
+    list_input_files : list[str]
+        list of centroided mzML filepaths from LC-MS metabolomics. 
+        Usually found in a folder.
+    parameters : dict
+        parameter dictionary passed from main.py, 
+        which imports from default_parameters and updates the dict by user arguments.
+
+    Outputs
+    -------
+    A local folder with asari extracted XICs (aka EICs or mass tracks), 
+    without full processing, in pickle files.
     '''
     sample_registry = register_samples(list_input_files)
     parameters['database_mode'] = 'ondisk'
@@ -218,17 +370,28 @@ def process_xics(list_input_files, parameters):
     create_export_folders(parameters, time_stamp)
     # samples are processed to mass tracks (EICs) here
     shared_dict = batch_EIC_from_samples_(sample_registry, parameters)
-    print("XICs were stored as pickle objects under %s" %os.path.join(parameters['outdir'], 'pickle'))
+    print("XICs were stored as pickle objects under %s" 
+          %os.path.join(parameters['outdir'], 'pickle'))
 
 # -----------------------------------------------------------------------------
-# main workflow for `extract`
+# main workflow for `extract`, short cut by passing mzlist to parameters['target']
 # -----------------------------------------------------------------------------
 
 def get_mz_list(infile):
     '''
-    Get a list of m/z valuies from infile, which is tab or comma delimited and has m/z in the first column.
-    Assuming first line as header and skip it.
+    Get a list of m/z valuies from infile, to be used as targets to extract from LC-MS features.
+    Currently, `extract` is a short cut function - asari still processes full feature tables, 
+    then searchs for the given targets.
+
+    Parameters
+    ----------
+    infile : str
+        filepath to input table, which is tab or comma delimited and has m/z in the first column,
+        header as first row.
+
+    Returns
+    -------
+    A list of m/z values.
     '''
     lines = open(infile).read().splitlines()[1:]
     return [float(x.split('\t')[0].split(",")[0]) for x in lines]
-
